import datetime
import io
from unittest.mock import Mock

import pytest
import sqlalchemy as sa
from click import progressbar
from click.testing import CliRunner

import datanommer.models as m
from datanommer.commands.extract_users import main as extract_users

from .utils import generate_bodhi_update_complete_message, generate_message


@pytest.fixture
def bodhi_message_db(datanommer_models):
    msg = generate_bodhi_update_complete_message()
    m.add(msg)
    m.session.execute(m.users_assoc_table.delete())
    msg_in_db = m.Message.from_msg_id(msg.id)
    msg_in_db.agent_name = None
    m.session.commit()

    m.session.refresh(msg_in_db)
    assert len(msg_in_db.users) == 0
    assert msg_in_db.agent_name is None
    return msg_in_db


@pytest.fixture(autouse=True)
def no_expunge(datanommer_models, monkeypatch):
    monkeypatch.setattr(m.session, "expunge_all", Mock(name="expunge_all"))
    monkeypatch.setattr(m.session, "expunge", Mock(name="expunge"))


def test_extract_users(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--debug", "usernames"])

    assert result.exit_code == 0, result.output
    expected_output = (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
        f"Usernames for message {bodhi_message_db.msg_id} of topic {bodhi_message_db.topic}: "
        "dudemcpants, ryanlerch\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
    )
    assert result.output == expected_output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) > 0
    assert {u.name for u in bodhi_message_db.users} == {"dudemcpants", "ryanlerch"}


def test_extract_users_topic(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(
        extract_users, ["--topic", "org.fedoraproject.stg.bodhi.update.comment", "usernames"]
    )

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) > 0
    assert {u.name for u in bodhi_message_db.users} == {"dudemcpants", "ryanlerch"}


def test_extract_users_wrong_topic(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--topic", "something.else", "usernames"])

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) == 0


def test_extract_users_category(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--category", "bodhi", "usernames"])

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) > 0
    assert {u.name for u in bodhi_message_db.users} == {"dudemcpants", "ryanlerch"}


def test_extract_users_wrong_category(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--category", "git", "usernames"])

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) == 0


def test_extract_users_topic_and_category(mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(
        extract_users, ["--category", "bodhi", "--topic", "some.topic", "usernames"]
    )
    assert result.exit_code != 0, result.output
    assert "Error: can't use both --topic and --category, choose one." in result.output


def test_extract_users_skipped_topic(bodhi_message_db, mock_config, mock_init):
    bodhi_message_db.topic = "org.release-monitoring.prod.anitya.project.version.update"
    m.session.commit()

    runner = CliRunner()
    result = runner.invoke(extract_users, ["usernames"])

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) == 0


def test_extract_users_no_users(datanommer_models, mock_config, mock_init):
    msg = generate_message()
    # change the schema header or the script won't pick it up
    msg._headers["fedora_messaging_schema"] = "testing"
    m.add(msg)
    runner = CliRunner()
    result = runner.invoke(extract_users, ["usernames"])

    assert result.exit_code == 0, result.output
    users_count = m.session.scalar(sa.select(sa.func.count(m.users_assoc_table.c.msg_id)))
    assert users_count == 0
    start = datetime.datetime.fromisoformat(msg._headers["sent-at"]).astimezone()
    start = str(start).split("+")[0]
    assert result.output == (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {start}\n"
        f"Working on 10000 messages sent after {start}\n"
    )


def test_extract_start(datanommer_models, mock_config, mock_init):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    msg = generate_bodhi_update_complete_message()
    # Set the message to have happenned 3 days ago
    msg._properties.headers["sent-at"] = (now - datetime.timedelta(days=3)).isoformat()
    m.add(msg)
    m.session.execute(m.users_assoc_table.delete())
    m.session.commit()

    runner = CliRunner()
    # Only look at messages from yesterday on
    result = runner.invoke(
        extract_users,
        ["--start", (now - datetime.timedelta(days=1)).strftime(r"%Y-%m-%d"), "usernames"],
    )

    assert result.exit_code == 0, result.output
    # Message must not have had users set
    users_count = m.session.scalar(sa.select(sa.func.count(m.users_assoc_table.c.msg_id)))
    assert users_count == 0
    assert result.output == "Counting messages...\nNo messages matched.\n"


def test_extract_end(bodhi_message_db, mock_config, mock_init):
    now = datetime.datetime.now()
    runner = CliRunner()
    # Only look at messages from yesterday on
    result = runner.invoke(
        extract_users,
        ["--end", (now - datetime.timedelta(days=1)).strftime(r"%Y-%m-%d"), "usernames"],
    )

    assert result.exit_code == 0, result.output
    # Message must not have had users set
    users_count = m.session.scalar(sa.select(sa.func.count(m.users_assoc_table.c.msg_id)))
    assert users_count == 0
    assert result.output == "Counting messages...\nNo messages matched.\n"


def test_extract_force_schema(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--force-schema", "base.message", "usernames"])

    assert result.exit_code == 0, result.output

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) == 0


def test_extract_invalid_message(bodhi_message_db, mock_config, mock_init):
    bodhi_message_db.msg = "this is invalid"
    m.session.commit()

    runner = CliRunner()
    result = runner.invoke(extract_users, ["usernames"])

    assert result.exit_code == 0, result.output
    assert result.output == (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
        f"Could not load message {bodhi_message_db.msg_id} on topic "
        f"{bodhi_message_db.topic}: 'this is invalid' is not of type 'object'\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
    )

    m.session.refresh(bodhi_message_db)
    assert len(bodhi_message_db.users) == 0


def test_extract_agent(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["agent"])

    assert result.exit_code == 0, result.output
    assert result.output == (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
    )
    m.session.refresh(bodhi_message_db)
    assert bodhi_message_db.agent_name == "dudemcpants"


def test_extract_agent_with(bodhi_message_db, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--debug", "agent"])

    assert result.exit_code == 0, result.output
    expected_output = (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
        f"Agent for message {bodhi_message_db.msg_id} of topic {bodhi_message_db.topic}: "
        "dudemcpants\n"
        f"Working on 10000 messages sent after {bodhi_message_db.timestamp}\n"
    )
    assert result.output == expected_output


def test_extract_agent_no_users(datanommer_models, mock_config, mock_init):
    msg = generate_message()
    # change the schema header or the script won't pick it up
    msg._headers["fedora_messaging_schema"] = "testing"
    m.add(msg)
    runner = CliRunner()
    result = runner.invoke(extract_users, ["agent"])

    assert result.exit_code == 0, result.output
    msg_in_db = m.Message.from_msg_id(msg.id)
    assert msg_in_db.agent_name is None
    assert result.output == (
        "Counting messages...\n"
        "Considering 1 message\n\n"
        f"Working on 10000 messages sent after {msg_in_db.timestamp}\n"
        f"Working on 10000 messages sent after {msg_in_db.timestamp}\n"
    )


def test_extract_is_tty(bodhi_message_db, mock_config, mock_init, mocker):
    output = io.StringIO()
    mocker.patch.object(output, "isatty", lambda: True)
    mocker.patch(
        "datanommer.commands.utils.click.progressbar", lambda **kw: progressbar(file=output, **kw)
    )
    runner = CliRunner()
    result = runner.invoke(extract_users, ["--debug", "usernames"])

    assert result.exit_code == 0, result.output
    expected_output = (
        "Counting messages...\n"
        "Considering 1 message\n"
        f"Usernames for message {bodhi_message_db.msg_id} of topic {bodhi_message_db.topic}: "
        "dudemcpants, ryanlerch\n"
    )
    assert result.output == expected_output



from bodhi.messages.schemas.update import UpdateCommentV1
from fedora_messaging import message as fedora_message


def generate_message(
    topic="org.fedoraproject.test.a.nice.message",
    body=None,
    headers=None,
):
    body = body or {"encouragement": "You're doing great!"}
    return fedora_message.Message(topic=topic, body=body, headers=headers)


def generate_bodhi_update_complete_message():
    msg = UpdateCommentV1(
        body={
            "comment": {
                "karma": -1,
                "text": "text",
                "timestamp": "2019-03-18 16:54:48",
                "update": {
                    "alias": "FEDORA-EPEL-2021-f2d195dada",
                    "builds": [
                        {"nvr": "abrt-addon-python3-2.1.11-50.el7"},
                        {"nvr": "kernel-10.4.0-2.el7"},
                    ],
                    "status": "pending",
                    "release": {"name": "F35"},
                    "request": "testing",
                    "user": {"name": "ryanlerch"},
                },
                "user": {"name": "dudemcpants"},
            }
        }
    )
    msg.topic = f"org.fedoraproject.stg.{msg.topic}"
    return msg



import pytest

import datanommer.commands


pytest_plugins = "datanommer.models.testing"


@pytest.fixture
def mock_init(mocker):
    # This is actually not very useful because init() checks a private attribute on the
    # session object to avoid being called twice. It just prevents a warning log.
    mocker.patch("datanommer.commands.m.init")


@pytest.fixture
def mock_config(mocker):
    mocker.patch.dict(
        datanommer.commands.utils.fedora_messaging_config.conf["consumer_config"],
        {
            "datanommer_sqlalchemy_url": "",
            "alembic_ini": None,
        },
    )



# This file is a part of datanommer, a message sink for fedmsg.
# Copyright (C) 2014, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import time
from datetime import datetime, timedelta

import pytest
from click import ClickException
from click.testing import CliRunner

import datanommer.commands
import datanommer.models as m

from .utils import generate_bodhi_update_complete_message, generate_message


def test_get_datanommer_sqlalchemy_url_keyerror(mocker):
    mocker.patch.dict(
        datanommer.commands.utils.fedora_messaging_config.conf["consumer_config"],
        {},
        clear=True,
    )
    with pytest.raises(ClickException):
        datanommer.commands.get_config()


def test_get_datanommer_sqlalchemy_url_config(mocker):
    conf = {
        "datanommer_sqlalchemy_url": "",
        "alembic_ini": "/some/where",
    }
    mocker.patch.dict(
        datanommer.commands.utils.fedora_messaging_config.conf["consumer_config"], conf
    )
    load_config = mocker.patch(
        "datanommer.commands.utils.fedora_messaging_config.conf.load_config",
    )
    datanommer.commands.get_config("some-path")
    load_config.assert_called_with("some-path")


def test_create(mocker):
    mock_model_init = mocker.patch("datanommer.commands.m.init")
    mocker.patch.dict(
        datanommer.commands.utils.fedora_messaging_config.conf["consumer_config"],
        {
            "datanommer_sqlalchemy_url": "TESTURL",
            "alembic_ini": "/some/where",
        },
    )

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.create, [])
    assert result.exit_code == 0, result.output

    assert result.output == "Creating Datanommer database and tables\n"
    mock_model_init.assert_called_once_with("TESTURL", alembic_ini="/some/where", create=True)


def test_stats(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.stats, [])
    assert result.exit_code == 0, result.output

    assert "git has 2 entries" in result.output
    assert "fas has 1 entries" in result.output


def test_stats_topics(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.stats, ["--topic"])
    assert result.exit_code == 0, result.output

    assert "org.fedoraproject.prod.git.receive.valgrind.master has 1 entries" in result.output
    assert "org.fedoraproject.stg.fas.user.create has 1 entries" in result.output
    assert "org.fedoraproject.prod.git.branch.valgrind.master has 1 entries" in result.output


def test_stats_category_topics(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.stats, ["--topic", "--category", "git"])
    assert result.exit_code == 0, result.output

    assert "org.fedoraproject.prod.git.receive.valgrind.master has 1 entries" in result.output
    assert "org.fedoraproject.stg.fas.user.create has 1 entries" not in result.output
    assert "org.fedoraproject.prod.git.branch.valgrind.master has 1 entries" in result.output


def test_stats_category(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.stats, ["--category", "git"])
    assert result.exit_code == 0, result.output

    assert result.output == "git has 2 entries\n"


def test_dump(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    m.add(msg2)

    msg3 = generate_bodhi_update_complete_message()
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.dump, [])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["topic"] == "org.fedoraproject.prod.git.branch.valgrind.master"


def test_dump_before(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    msg1._properties.headers["sent-at"] = datetime(2013, 2, 14).isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    msg2._properties.headers["sent-at"] = datetime(2013, 2, 15).isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.log.receive.valgrind.master")
    msg3._properties.headers["sent-at"] = datetime(2013, 2, 16, 8).isoformat()
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.dump, ["--before", "2013-02-16"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["topic"] == "org.fedoraproject.prod.git.branch.valgrind.master"
    assert json_object[1]["topic"] == "org.fedoraproject.prod.git.receive.valgrind.master"
    assert len(json_object) == 2


def test_dump_since(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    msg1._properties.headers["sent-at"] = datetime(2013, 2, 14).isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    msg2._properties.headers["sent-at"] = datetime(2013, 2, 15).isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.log.receive.valgrind.master")
    msg3._properties.headers["sent-at"] = datetime(2013, 2, 16, 8).isoformat()
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.dump, ["--since", "2013-02-14T08:00:00"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["topic"] == "org.fedoraproject.prod.git.receive.valgrind.master"
    assert json_object[1]["topic"] == "org.fedoraproject.prod.log.receive.valgrind.master"
    assert len(json_object) == 2


def test_dump_timespan(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    msg1._properties.headers["sent-at"] = datetime(2013, 2, 14).isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    msg2._properties.headers["sent-at"] = datetime(2013, 2, 15).isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.log.receive.valgrind.master")
    msg3._properties.headers["sent-at"] = datetime(2013, 2, 16, 8).isoformat()
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(
        datanommer.commands.dump,
        ["--before", "2013-02-16", "--since", "2013-02-14T08:00:00"],
    )
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["topic"] == "org.fedoraproject.prod.git.receive.valgrind.master"
    assert len(json_object) == 1


def test_dump_invalid_dates(datanommer_models, mock_config, mock_init):
    runner = CliRunner()
    result = runner.invoke(datanommer.commands.dump, ["--before", "2013-02-16asdasd"])
    assert result.exit_code > 0, result.output
    assert result.output == "Error: Invalid date format\n"

    result = runner.invoke(datanommer.commands.dump, ["--since", "2013-02-16asdasd"])
    assert result.exit_code > 0, result.output
    assert result.output == "Error: Invalid date format\n"


def test_latest_overall(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--overall"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["git"]["body"] == {"Message 3": "Message 3"}
    assert len(json_object) == 1


def test_latest_topic(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(
        datanommer.commands.latest, ["--topic", "org.fedoraproject.stg.fas.user.create"]
    )
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["fas"]["body"] == {"Message 2": "Message 2"}
    assert len(json_object) == 1


def test_latest_category(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--category", "fas"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[0]["fas"]["body"] == {"Message 2": "Message 2"}
    assert len(json_object) == 1


def test_latest_timestamp_human(datanommer_models, mocker, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    msg1._properties.headers["sent-at"] = datetime(2013, 2, 14).isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.stg.fas.user.create")
    msg2._properties.headers["sent-at"] = datetime(2013, 2, 15, 15, 15, 15, 15).isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    msg3._properties.headers["sent-at"] = datetime(2013, 2, 16, 16, 16, 16, 16).isoformat()
    m.add(msg3)

    # datanommer-latest defaults to the last year, so mock the
    # datetime calls to go back to 2013
    mock_dt = mocker.patch("datanommer.commands.datetime")
    mock_dt.now.return_value = datetime(2013, 3, 1)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--timestamp", "--human"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[1] == "2013-02-16 16:16:16.000016"
    assert json_object[0] == "2013-02-15 15:15:15.000015"
    assert len(json_object) == 2


def test_latest_timestamp(datanommer_models, mocker, mock_config, mock_init):
    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    msg1._properties.headers["sent-at"] = datetime(2013, 2, 14).isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.stg.fas.user.create")
    msg2._properties.headers["sent-at"] = datetime(2013, 2, 15).isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    msg3._properties.headers["sent-at"] = datetime(2013, 2, 16).isoformat()
    m.add(msg3)

    # datanommer-latest defaults to the last year, so mock the
    # datetime calls to go back to 2013
    mock_dt = mocker.patch("datanommer.commands.datetime")
    mock_dt.now.return_value = datetime(2013, 3, 1)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--timestamp"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[1] == time.mktime(datetime(2013, 2, 16).timetuple())
    assert json_object[0] == time.mktime(datetime(2013, 2, 15).timetuple())
    assert len(json_object) == 2


def test_latest_timesince(datanommer_models, mocker, mock_config, mock_init):
    now = datetime(2013, 3, 1)

    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    time1 = now - timedelta(days=1)
    msg1._properties.headers["sent-at"] = time1.isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.stg.fas.user.create")
    time2 = now - timedelta(seconds=60)
    msg2._properties.headers["sent-at"] = time2.isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    time3 = now - timedelta(seconds=1)
    msg3._properties.headers["sent-at"] = time3.isoformat()
    m.add(msg3)

    # datanommer-latest defaults to the last year, so mock the
    # datetime calls to go back to 2013
    mock_dt = mocker.patch("datanommer.commands.datetime")
    mock_dt.now.return_value = now

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--timesince"])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    # allow .1 second to run test
    assert int(json_object[1]) <= 1.1
    assert int(json_object[1]) >= 1
    assert int(json_object[0]) <= 60.1
    assert int(json_object[0]) >= 60
    assert len(json_object) == 2


def test_latest_timesince_human(datanommer_models, mock_config, mock_init, mocker):
    now = datetime.now()
    # mocker.patch.object(datanommer.commands.datetime, "now", return_value=now)
    patched_datetime = mocker.patch("datanommer.commands.datetime", mocker.Mock(wraps=datetime))
    patched_datetime.now.return_value = now

    msg1 = generate_message(topic="org.fedoraproject.prod.git.branch.valgrind.master")
    time1 = now - timedelta(days=2)
    msg1._properties.headers["sent-at"] = time1.isoformat()
    m.add(msg1)

    msg2 = generate_message(topic="org.fedoraproject.stg.fas.user.create")
    time2 = now - timedelta(days=1)
    msg2._properties.headers["sent-at"] = time2.isoformat()
    m.add(msg2)

    msg3 = generate_message(topic="org.fedoraproject.prod.git.receive.valgrind.master")
    time3 = now - timedelta(seconds=1)
    msg3._properties.headers["sent-at"] = time3.isoformat()
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, ["--timesince", "--human"])
    assert result.exit_code == 0, result.output

    assert json.loads(result.output) == ["1 day, 0:00:00", "0:00:01"]


def test_latest(datanommer_models, mock_config, mock_init):
    msg1 = generate_message(
        topic="org.fedoraproject.prod.git.branch.valgrind.master",
        body={"Message 1": "Message 1"},
    )
    time1 = datetime.now() - timedelta(days=2)
    msg1._properties.headers["sent-at"] = time1.isoformat()
    m.add(msg1)

    msg2 = generate_message(
        topic="org.fedoraproject.stg.fas.user.create", body={"Message 2": "Message 2"}
    )
    m.add(msg2)

    msg3 = generate_message(
        topic="org.fedoraproject.prod.git.receive.valgrind.master",
        body={"Message 3": "Message 3"},
    )
    m.add(msg3)

    runner = CliRunner()
    result = runner.invoke(datanommer.commands.latest, [])
    assert result.exit_code == 0, result.output

    json_object = json.loads(result.output)

    assert json_object[1]["git"]["body"] == {"Message 3": "Message 3"}
    assert json_object[0]["fas"]["body"] == {"Message 2": "Message 2"}
    assert len(json_object) == 2



import logging

import click
from fedora_messaging import config as fedora_messaging_config
from fedora_messaging.message import load_message as load_message
from sqlalchemy import func

import datanommer.models as m


# Go trough messages these many at a time
CHUNK_SIZE = 10000
log = logging.getLogger(__name__)


def get_config(config_path=None):
    if config_path:
        fedora_messaging_config.conf.load_config(config_path)
    conf = fedora_messaging_config.conf["consumer_config"]
    for key in ("datanommer_sqlalchemy_url", "alembic_ini"):
        if key not in conf:
            raise click.ClickException(f"{key} not defined in the fedora-messaging config")
    return conf


config_option = click.option(
    "-c",
    "--config",
    "config_path",
    help="Load this Fedora Messaging config file",
    type=click.Path(exists=True, readable=True),
)


def iterate_over_messages(query, start, chunk_size):
    click.echo("Counting messages...")

    total = m.session.scalar(query.with_only_columns(func.count(m.Message.id)))
    if not total:
        click.echo("No messages matched.")
        return

    click.echo(f"Considering {total} message{'s' if total > 1 else ''}")

    query = query.order_by(m.Message.timestamp)
    with click.progressbar(length=total) as bar:
        has_messages = True
        chunk_start = start
        first_run = True
        while has_messages:
            # click < 8.2 (Python < 3.10): use bar.is_hidden
            # click >= 8.2 (Python >= 3.10): use bar.hidden and the TTY check
            if (hasattr(bar, "is_hidden") and bar.is_hidden) or (
                hasattr(bar, "hidden") and (bar.hidden or not bar.file.isatty())
            ):
                click.echo(f"Working on {chunk_size} messages sent after {chunk_start}")
            chunk_query = query.where(m.Message.timestamp >= chunk_start).limit(chunk_size)
            if not first_run:
                chunk_query = chunk_query.offset(1)
            has_messages = False
            for message in m.session.scalars(chunk_query):
                bar.update(1)
                has_messages = True
                yield message
            if has_messages:
                chunk_start = message.timestamp
            first_run = False
            m.session.commit()
            m.session.expunge_all()



# This file is a part of datanommer, a message sink for fedmsg.
# Copyright (C) 2014, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
import importlib.metadata
import itertools
import json
import logging
import time
from datetime import datetime, timedelta, timezone

import click
from sqlalchemy import func, select

import datanommer.models as m

from .utils import config_option, get_config


__version__ = importlib.metadata.version("datanommer-commands")

log = logging.getLogger("datanommer")


@click.command()
@config_option
def create(config_path):
    """Create a database and tables for 'datanommer.sqlalchemy.url'"""
    config = get_config(config_path)
    click.echo("Creating Datanommer database and tables")
    m.init(
        config["datanommer_sqlalchemy_url"],
        alembic_ini=config["alembic_ini"],
        create=True,
    )


@click.command()
@config_option
@click.option("--since", default=None, help="Only after datetime, ex 2013-02-14T08:05:59.87")
@click.option("--before", default=None, help="Only before datetime, ex 2013-02-14T08:05:59.87")
def dump(config_path, since, before):
    """Dump the contents of the datanommer database as JSON.

    You can also specify a timespan with the --since and --before arguments:

        $ datanommer-dump --before 2013-02-15 --since 2013-02-11T08:00:00 > datanommer-dump.json
    """
    config = get_config(config_path)
    m.init(
        config["datanommer_sqlalchemy_url"],
        alembic_ini=config["alembic_ini"],
    )

    query = select(m.Message)
    if before:
        try:
            before = datetime.fromisoformat(before)
        except ValueError as e:
            raise click.ClickException("Invalid date format") from e

        query = query.where(m.Message.timestamp <= before)

    if since:
        try:
            since = datetime.fromisoformat(since)
        except ValueError as e:
            raise click.ClickException("Invalid date format") from e

        query = query.where(m.Message.timestamp >= since)

    results = [json.dumps(msg.as_fedora_message_dict()) for msg in m.session.scalars(query)]
    click.echo(f"[{','.join(results)}]")


@click.command()
@config_option
@click.option("--topic", is_flag=True, help="Shows the stats per topic")
@click.option(
    "--category",
    default=None,
    help="Shows the stats within only the specified category",
)
def stats(config_path, topic, category):
    """Produce stats on the contents of the datanommer database.

    The default is to display the stats per category. You can also display
    the stats per topic with the --topic argument:

        $ datanommer-stats --topic
        org.fedoraproject.stg.fas.group.member.remove has 10 entries
        org.fedoraproject.stg.logger.log has 76 entries
        org.fedoraproject.stg.bodhi.update.comment has 5 entries
        org.fedoraproject.stg.busmon.colorized-messages has 10 entries
        org.fedoraproject.stg.fas.user.update has 10 entries
        org.fedoraproject.stg.wiki.article.edit has 106 entries
        org.fedoraproject.stg.fas.user.create has 3 entries
        org.fedoraproject.stg.bodhitest.testing has 4 entries
        org.fedoraproject.stg.fedoratagger.tag.create has 9 entries
        org.fedoraproject.stg.fedoratagger.user.rank.update has 5 entries
        org.fedoraproject.stg.wiki.upload.complete has 1 entries
        org.fedoraproject.stg.fas.group.member.sponsor has 6 entries
        org.fedoraproject.stg.fedoratagger.tag.update has 1 entries
        org.fedoraproject.stg.fas.group.member.apply has 17 entries
        org.fedoraproject.stg.__main__.testing has 1 entries

    The --category argument can be combined with --topic to shows stats of the
    topics with a specific category or can be used alone to show the stats for
    only the one category:

        $ datanommer-stats --topic --category fas
        org.fedoraproject.stg.fas.group.member.remove has 10 entries
        org.fedoraproject.stg.fas.user.update has 10 entries
        org.fedoraproject.stg.fas.user.create has 3 entries
        org.fedoraproject.stg.fas.group.member.sponsor has 6 entries
        org.fedoraproject.stg.fas.group.member.apply has 17 entries

        $ datanommmer-stats --category fas
        fas has 46 entries

    """
    config = get_config(config_path)
    m.init(
        config["datanommer_sqlalchemy_url"],
        alembic_ini=config["alembic_ini"],
    )

    if topic:
        query = select(m.Message.topic, func.count(m.Message.topic))
        if category:
            query = query.where(m.Message.category == category)
        query = query.group_by(m.Message.topic)
    else:
        query = select(m.Message.category, func.count(m.Message.category))
        if category:
            query = query.where(m.Message.category == category)
        query = query.group_by(m.Message.category)

    results = m.session.execute(query).all()

    if topic:
        for topic, count in results:
            click.echo(f"{topic} has {count} entries")
    else:
        for category, count in results:
            click.echo(f"{category} has {count} entries")


@click.command()
@config_option
@click.option("--topic", default=None, help="Show the latest for only a specific topic.")
@click.option("--category", default=None, help="Show the latest for only a specific category.")
@click.option(
    "--overall",
    is_flag=True,
    help="Show only the latest message out of all message types.",
)
@click.option("--timestamp", is_flag=True, help="Show only the timestamp of the message(s).")
@click.option(
    "--timesince",
    is_flag=True,
    help="Show the number of seconds since the last message",
)
@click.option(
    "--human",
    is_flag=True,
    help="When combined with --timestamp or --timesince,show a human readable date.",
)
def latest(config_path, topic, category, overall, timestamp, timesince, human):
    """Print the latest message(s) ingested by datanommer.

    The default is to display the latest message in each message category. The
    latest in only a specified category or topic can also be returned::

        $ datanommer-latest --category bodhi
        [{"bodhi": {
          "topic": "org.fedoraproject.stg.bodhi.update.comment",
          "msg": {
            "comment": {
              "group": null,
              "author": "ralph",
              "text": "Testing for latest datanommer.",
              "karma": 0,
              "anonymous": false,
              "timestamp": 1360349639.0,
              "update_title": "xmonad-0.10-10.fc17"
            },
            "agent": "ralph"
          },
        }}]

        $ datanommer-latest --topic org.fedoraproject.stg.bodhi.update.comment
        [{"bodhi": {
          "topic": "org.fedoraproject.stg.bodhi.update.comment",
          "msg": {
            "comment": {
              "group": null,
              "author": "ralph",
              "text": "Testing for latest datanommer.",
              "karma": 0,
              "anonymous": false,
              "timestamp": 1360349639.0,
              "update_title": "xmonad-0.10-10.fc17"
            },
            "agent": "ralph"
          },
        }}]

    Or to display the latest, regardless of the topic or category::

        $ datanommer-latest --overall
        [{"bodhi": {
          "topic": "org.fedoraproject.stg.bodhi.update.comment",
          "msg": {
            "comment": {
              "group": null,
              "author": "ralph",
              "text": "Testing for latest datanommer.",
              "karma": 0,
              "anonymous": false,
              "timestamp": 1360349639.0,
              "update_title": "xmonad-0.10-10.fc17"
            },
            "agent": "ralph"
          },
        }}]

    You can combine either a --topic, --category or --overall argument while
    requesting information about the timestamp of the latest::

        $ datanommer-latest --category wiki --timestamp
        [1361166918.0]

        # February 18, 2013 at 5:55AM
        $ datanommer-latest --category wiki --timestamp --human
        ["2013-02-18 05:55:18"]

    Or how recent that timestamp is::

        # 49250 seconds ago
        $ datanommer-latest --category wiki --timesince
        [49250]

        # 13 hours, 40 minutes, 59.52 seconds ago
        $ datanommer-latest --category wiki --timesince --human
        [13:40:59.519447]
    """
    config = get_config(config_path)
    m.init(
        config["datanommer_sqlalchemy_url"],
        alembic_ini=config["alembic_ini"],
    )

    if topic:
        queries = [select(m.Message).where(m.Message.topic == topic)]

    elif category:
        queries = [select(m.Message).where(m.Message.category == category)]
    elif not overall:
        # If no args..
        categories_query = select(m.Message.category).distinct().order_by(m.Message.category)
        categories = m.session.scalars(categories_query)
        queries = [
            select(m.Message).where(m.Message.category == category) for category in categories
        ]
    else:
        # Show only the single latest message, regardless of type.
        queries = [select(m.Message)]

    # Only check messages from the last year to speed up queries
    a_year = timedelta(days=365)
    earliest = datetime.now(tz=timezone.utc) - a_year
    queries = [q.where(m.Message.timestamp > earliest) for q in queries]

    # Order and limit to the latest.
    queries = [q.order_by(m.Message.timestamp.desc()).limit(1) for q in queries]

    def formatter(key, val):
        if timestamp and human:
            return json.dumps(str(val.timestamp))
        elif timestamp:
            return json.dumps(time.mktime(val.timestamp.timetuple()))
        elif timesince and human:
            return json.dumps(str(datetime.now() - val.timestamp))
        elif timesince:
            timedelta = datetime.now() - val.timestamp
            return json.dumps(str((timedelta.days * 86400) + timedelta.seconds))
        else:
            return f'{{"{key}": {json.dumps(val.as_fedora_message_dict())}}}'

    results = []
    for result in itertools.chain.from_iterable(m.session.scalars(query) for query in queries):
        results.append(formatter(result.category, result))

    click.echo(f"[{','.join(results)}]")



import datetime
import logging

import click
from fedora_messaging.exceptions import ValidationError
from fedora_messaging.message import load_message as load_message
from sqlalchemy import and_, not_, select

import datanommer.models as m

from .utils import CHUNK_SIZE, config_option, get_config, iterate_over_messages


log = logging.getLogger(__name__)

USERNAMES_SKIP_TOPICS = [
    "%.anitya.%",
    "%.discourse.%",
    "%.hotness.update.bug.file",
    "%.hotness.update.drop",
    "%.koschei.%",
    "%.mdapi.%",
]
AGENT_SKIP_TOPICS = [
    "%.hotness.update.bug.file",
    "%.hotness.update.drop",
]


@click.group()
@config_option
@click.option("--topic", default=None, help="Only extract users for messages of a specific topic.")
@click.option(
    "--category",
    default=None,
    help="Only extract users for messages of a specific category.",
)
@click.option(
    "--start",
    default=None,
    type=click.DateTime(),
    help="Only extract users for messages after a specific timestamp.",
)
@click.option(
    "--end",
    default=None,
    type=click.DateTime(),
    help="Only extract users for messages before a specific timestamp.",
)
@click.option(
    "--force-schema",
    default=None,
    help=(
        "Force usage of this schema name to extract usernames. This is the key in the "
        "exposed entry point / plugin, for example: wiki.article.edit.v1"
    ),
)
@click.option(
    "--chunk-size",
    default=CHUNK_SIZE,
    type=int,
    show_default=True,
    help="Go through messages these many at a time (lower is slower but saves memory).",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Show more information.",
)
@click.pass_context
def main(ctx, config_path, topic, category, start, end, force_schema, chunk_size, debug):
    ctx.ensure_object(dict)
    ctx.obj["options"] = ctx.params
    ctx.obj["config"] = config = get_config(config_path)
    m.init(
        config["datanommer_sqlalchemy_url"],
        alembic_ini=config["alembic_ini"],
    )
    if topic and category:
        raise click.UsageError("can't use both --topic and --category, choose one.")

    if not start:
        ctx.obj["options"]["start"] = m.session.execute(
            select(m.Message.timestamp).order_by(m.Message.timestamp).limit(1)
        ).scalar_one()

    query = select(m.Message)
    if topic:
        query = query.where(m.Message.topic == topic)
    elif category:
        query = query.where(m.Message.category == category)

    query = query.where(m.Message.timestamp >= ctx.obj["options"]["start"])
    if end:
        query = query.where(m.Message.timestamp < end)
    else:
        end = datetime.datetime.now()

    if force_schema is None:
        query = query.where(
            m.Message.headers.has_key("fedora_messaging_schema"),
            m.Message.headers["fedora_messaging_schema"].astext != "base.message",
        )
    ctx.obj["query"] = query


@main.command("usernames")
@click.pass_context
def extract_usernames(ctx):
    """Go over old messages, extract users and store them.

    This is useful when a message schema has been added and we want to populate the users table
    with the new information.
    """
    debug = ctx.obj["options"]["debug"]
    query = ctx.obj["query"]
    query = query.where(
        and_(*[not_(m.Message.topic.like(skipped)) for skipped in USERNAMES_SKIP_TOPICS])
    )
    query = query.join(
        m.users_assoc_table,
        and_(
            m.Message.id == m.users_assoc_table.c.msg_id,
            m.Message.timestamp == m.users_assoc_table.c.msg_timestamp,
        ),
        isouter=True,
    ).where(m.users_assoc_table.c.msg_id.is_(None))

    for message in iterate_over_messages(
        query, ctx.obj["options"]["start"], ctx.obj["options"]["chunk_size"]
    ):
        fm_message = get_fedora_message(message, force_schema=ctx.obj["options"]["force_schema"])
        if fm_message is None or not fm_message.usernames:
            m.session.expunge(message)
            continue
        message._insert_list(m.User, m.users_assoc_table, fm_message.usernames)
        if debug:
            click.echo(
                f"Usernames for message {message.msg_id} of topic {message.topic}"
                f": {', '.join(fm_message.usernames)}"
            )


def get_fedora_message(db_message, force_schema):
    headers = db_message.headers
    if force_schema and headers is not None:
        headers["fedora_messaging_schema"] = force_schema
    try:
        fm_message = load_message(
            {
                "topic": db_message.topic,
                "headers": headers,
                "id": db_message.msg_id,
                "body": db_message.msg,
            }
        )
    except ValidationError as e:
        try:
            # Remove this block after fedora-messaging 3.6.0 and use e.summary
            error_msg = e.args[0].summary
        except AttributeError:
            error_msg = str(e).split("\n")[0]
        click.echo(
            f"Could not load message {db_message.msg_id} on topic {db_message.topic}: {error_msg}",
            err=True,
        )
        return None

    return fm_message


@main.command("agent")
@click.pass_context
def extract_agent(ctx):
    """Go over old messages, extract the agent_name and store it.

    This is useful when a message schema has been added and we want to populate the agent_name
    column with the new information.
    """
    debug = ctx.obj["options"]["debug"]
    query = ctx.obj["query"]
    query = query.where(
        and_(*[not_(m.Message.topic.like(skipped)) for skipped in AGENT_SKIP_TOPICS])
    )
    query = query.where(m.Message.agent_name.is_(None))

    for message in iterate_over_messages(
        query, ctx.obj["options"]["start"], ctx.obj["options"]["chunk_size"]
    ):
        fm_message = get_fedora_message(message, force_schema=ctx.obj["options"]["force_schema"])
        if fm_message is None or not fm_message.agent_name:
            m.session.expunge(message)
            continue
        message.agent_name = fm_message.agent_name
        if debug:
            click.echo(
                f"Agent for message {message.msg_id} of topic {message.topic}"
                f": {fm_message.agent_name}"
            )
